/*
 * Segment 5 Intelligente Analyseverfahren, SW-Einheit 5.2 Straßensubsegmentanalyse
 * Copyright (C) 2007-2018 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.subsegana.subsegana.application;

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.iav.fuzzylib.fuzzylib.*;
import de.bsvrz.iav.subsegana.subsegana.domain.*;
import de.bsvrz.iav.subsegana.subsegana.domain.shared.*;
import de.bsvrz.iav.subsegana.subsegana.domain.verkehrsdaten.*;
import de.bsvrz.iav.subsegana.subsegana.portals.*;
import de.bsvrz.iav.subsegana.subsegana.shared.*;
import de.bsvrz.sys.funclib.application.*;
import de.bsvrz.sys.funclib.commandLineArgs.*;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList.*;
import de.bsvrz.sys.funclib.debug.*;

import java.util.*;
import java.util.stream.*;

/**
 * Main-Klasse der Straßensubsegmentanalyse.
 * <p>
 * Die folgende Abbildung zeigt den Datenfluss der Straßensubsegmentanalyse.
 * <p>
 * <img src="doc-files/strassensubsegmentanalyse-datenfluss.png" alt="Datenfluss">
 * <p>
 * Die Funktionseinheiten in der Abbildung unterteilen sich in Kreise für Domainobjekte, Rechtecke für Portale
 * (Interaktion) und Dreiecke für Provider (Ressourcen).
 * <p>
 * Die Aggregation puffert aktuelle Messwerte für den aktuellen Zyklus, bis diese Messwerte aggregiert werden
 * können. Dazu kennt die Aggregation die Messquerschnitte, die zu einem Straßenteilsegment gehören und publiziert einen
 * über diese Messquerschnitte aggregierten Messwert für das dazugehörige Straßenteilsegment.
 * <p>
 * Die Fuzzyfizierung hält die Fuzzy-Sets der verwalteten Objekte vor, für die Messwerte fuzzyfiziert werden.
 * <p>
 * Die Inferenz puffert die Fuzzy-Werte der letzten Zyklen, da in Regeln nicht nur auf aktuelle, sondern auch auf
 * vergangene Fuzzy-Werte zugegriffen werden kann. So kann in einer Regel zum Beispiel der aktuelle mit dem letzten
 * Fuzzy-Wert kombiniert werden. Zu dem hält die Inferenz die Regelbasen der verwalteten Objekte vor, für die Regelbasen
 * ausgewertet werden.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
@SuppressWarnings("ClassDataAbstractionCoupling")
public class Strassensubsegmentanalyse implements StandardApplication {

    private final Event<Collection<Datensatz<?, FuzzyVariable>>> fuzzyWertErzeugt = new Event<>();

    private final Event<Collection<Datensatz<?, Stoerfallzustand>>> verkehrszustandErzeugt = new Event<>();

    private Debug log;

    private Collection<String> argSystemobjekte;
    private int argMaxZyklen;

    private Aggregation aggregation;
    private Fuzzyfizierung fuzzyfizierung;
    private Inferenz inferenz;
    private Verkehrszustandsbestimmung verkehrszustandsbestimmung;

    private MesswertEmpfaenger messwertEmpfaenger;
    private ParameterEmpfaenger parameterEmpfaenger;
    private FuzzyWertSender fuzzyWertSender;
    private VerkehrszustandSender verkehrszustandSender;

    private Strassensubsegmentanalyse() {
        // Konstruktor verstecken
    }

    /**
     * Test-Seam.
     */
    Strassensubsegmentanalyse(int maxZyklen) {
        log = Debug.getLogger();
        initializeDomain(maxZyklen);
    }

    public static void main(final String[] args) {
        StandardApplicationRunner.run(new Strassensubsegmentanalyse(), args);
    }

    @Override
    public void parseArguments(final ArgumentList argumentList) {
        log = Debug.getLogger();
        argSystemobjekte = bestimmeZuVerarbeitendeSystemobjekte(argumentList);
        argMaxZyklen = bestimmeMaximaleAnzahlZuCachenderZyklen(argumentList);
    }

    private static Collection<String> bestimmeZuVerarbeitendeSystemobjekte(ArgumentList argumentList) {
        Argument arg = argumentList.fetchArgument("-objekte=");
        if (arg.asString().trim().isEmpty()) {
            return Collections.emptyList();
        }

        String[] pids = arg.asString().split(",");
        return Arrays.stream(pids)
                .map(String::trim)
                .distinct()
                .collect(Collectors.toList());
    }

    private static int bestimmeMaximaleAnzahlZuCachenderZyklen(ArgumentList argumentList) {
        Argument arg = argumentList.fetchArgument("-maxZyklen=2");
        return arg.intValueBetween(1, 9);
    }

    @Override
    public void initialize(final ClientDavInterface connection) {
        DataModel model = connection.getDataModel();
        ObjektFactory objektFactory = new ObjektFactory(model);
        Collection<Systemobjekt> objekte = objektFactory.bestimmeObjekte(argSystemobjekte);

        initializeDomain(argMaxZyklen);
        fuzzyWertSender = new FuzzyWertSender(connection);
        verkehrszustandSender = new VerkehrszustandSender(connection);
        parameterEmpfaenger = new ParameterEmpfaenger(connection, objektFactory);
        messwertEmpfaenger = new MesswertEmpfaenger(connection, objektFactory);

        parameterEmpfaenger.fuzzySetsEmpfangen().addHandler(this::fuzzySetsAktualisiert);
        parameterEmpfaenger.regelbasisEmpfangen().addHandler(this::regelbasisAktualisiert);
        messwertEmpfaenger.messwertEmpfangen().addHandler(this::verarbeiteMesswerte);
        this.fuzzyWertErzeugt.addHandler(fuzzyWertSender::sende);
        this.verkehrszustandErzeugt.addHandler(verkehrszustandSender::sende);

        fuzzyWertSender.anmelden(objekte);
        log.info("Fuzzy-Wert-Sender bereit.");

        verkehrszustandSender.anmelden(objekte);
        log.info("Verkehrszustandssender bereit.");

        parameterEmpfaenger.anmelden(objekte);
        log.info("Parameterempfänger bereit.");

        messwertEmpfaenger.anmelden(objekte);
        log.info("Messwertempfänger bereit.");

        log.info("Straßensubsegmentanalyse bereit.");
    }

    private void initializeDomain(int maxZyklen) {
        aggregation = new Aggregation();
        fuzzyfizierung = new Fuzzyfizierung();
        inferenz = new Inferenz(maxZyklen);
        verkehrszustandsbestimmung = new Verkehrszustandsbestimmung();

        aggregation.messwertAggregiert().addHandler(fuzzyfizierung::fuzzyfiziere);
        log.info("Aggregation bereit.");

        fuzzyfizierung.messwerteFuzzyfiziert().addHandler(fuzzyWertErzeugt::send);
        fuzzyfizierung.messwerteFuzzyfiziert().addHandler(inferenz::werteRegelnAus);
        log.info("Fuzzyfizierung bereit.");

        inferenz.fuzzyWerteAbgeleitet().addHandler(fuzzyWertErzeugt::send);
        inferenz.fuzzyWerteAbgeleitet().addHandler(verkehrszustandsbestimmung::bestimmeVerkehrszustand);
        log.info("Inferenz bereit.");

        verkehrszustandsbestimmung.verkehrszustandBestimmt().addHandler(verkehrszustandErzeugt::send);
        log.info("Verkehrszustandsbestimmung bereit.");
    }

    /**
     * Test-Seam.
     */
    void fuzzySetsAktualisiert(Datensatz<?, LinguistischeVariable> parameter) {
        fuzzyfizierung.fuzzySetsAktualisiert(parameter);
    }

    /**
     * Test-Seam.
     */
    void regelbasisAktualisiert(Datensatz<?, Regelbasis> parameter) {
        inferenz.regelbasisAktualisiert(parameter);
    }

    /**
     * Test-Seam.
     */
    void verarbeiteMesswerte(Datensatz<?, Collection<Messwert>> messwerte) {
        if (messwerte.getSystemobjekt() instanceof Messquerschnitt) {
            aggregation.aggregiereMesswert((Datensatz<Messquerschnitt, Collection<Messwert>>) messwerte);
        }
        fuzzyfizierung.fuzzyfiziere(messwerte);
    }

    /**
     * Test-Seam.
     */
    void verarbeiteFuzzyWerte(Collection<Datensatz<?, FuzzyVariable>> fuzzyWerte) {
        inferenz.werteRegelnAus(fuzzyWerte);
    }

    /**
     * Test-Seam.
     */
    Event<Collection<Datensatz<?, FuzzyVariable>>> fuzzyWertErzeugt() {
        return fuzzyWertErzeugt;
    }

    /**
     * Test-Seam.
     */
    Event<Collection<Datensatz<?, Stoerfallzustand>>> verkehrszustandErzeugt() {
        return verkehrszustandErzeugt;
    }

}
