/*
 * Segment 5 Intelligente Analyseverfahren, SW-Einheit 5.2 Straßensubsegmentanalyse
 * Copyright (C) 2007-2018 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.subsegana.subsegana.portals;

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.iav.fuzzylib.fuzzylib.*;
import de.bsvrz.iav.subsegana.subsegana.domain.shared.*;
import de.bsvrz.iav.subsegana.subsegana.domain.umfelddaten.*;
import de.bsvrz.iav.subsegana.subsegana.domain.verkehrsdaten.*;
import de.bsvrz.iav.subsegana.subsegana.shared.*;
import de.bsvrz.sys.funclib.debug.*;
import de.bsvrz.sys.funclib.operatingMessage.*;

import java.text.*;
import java.util.*;
import java.util.stream.*;

import static de.bsvrz.iav.subsegana.subsegana.portals.Konstanten.*;

@SuppressWarnings("ClassFanOutComplexity")
class RegelbasisEmpfaenger extends Kommunikation {

    private final Event<Datensatz<?, Regelbasis>> regelbasisEmpfangen = new Event<>();

    private final Debug log = Debug.getLogger();
    private final MessageSender messageSender = MessageSender.getInstance();
    private final ClientReceiverInterface receiver = new Receiver();
    private final ObjektFactory objektFactory;

    RegelbasisEmpfaenger(ClientDavInterface verbindung, ObjektFactory objektFactory) {
        super(verbindung);
        this.objektFactory = objektFactory;
    }

    void anmelden(Collection<Systemobjekt> objekte) {
        DataDescription dd = datenbeschreibung(ATG_REGELBASIS, "asp.parameterSoll");

        Collection<SystemObject> messquerschnitte = konvertiereObjekte(objekte, obj -> obj instanceof Messquerschnitt);
        getVerbindung().subscribeReceiver(receiver, messquerschnitte, dd, ReceiveOptions.normal(), ReceiverRole.receiver());
        log.config(MessageFormat.format("Als Empfänger von Regelbasis für {0} Messquerschnitten angemeldet.", messquerschnitte.size()));

        Collection<SystemObject> strassenteilsegmente = konvertiereObjekte(objekte, obj -> obj instanceof Strassenteilsegment);
        getVerbindung().subscribeReceiver(receiver, strassenteilsegmente, dd, ReceiveOptions.normal(), ReceiverRole.receiver());
        log.config(MessageFormat.format("Als Empfänger von Regelbasis für {0} Straßenteilsegmente angemeldet.", strassenteilsegmente.size()));

        Collection<SystemObject> umfelddatenmessstellen = konvertiereObjekte(objekte, obj -> obj instanceof Umfelddatenmessstelle);
        getVerbindung().subscribeReceiver(receiver, umfelddatenmessstellen, dd, ReceiveOptions.normal(), ReceiverRole.receiver());
        log.config(MessageFormat.format("Als Empfänger von Regelbasis für {0} Umfelddatenmessstellen angemeldet.", umfelddatenmessstellen.size()));
    }

    Datensatz<?, Regelbasis> konvertiereDatensatz(ResultData result) {
        Systemobjekt objekt = konvertiereObjekt(result.getObject());
        Regelbasis regelbasis = konvertiereRegelbasis(result);
        Datensatz<Systemobjekt, Regelbasis> datensatz = new Datensatz<>(konvertiereZeitstempel(result.getDataTime()), objekt, regelbasis);
        if (regelbasis.nonNull()) {
            log.fine("Parameter für Regelbasis empfangen", datensatz);
        }
        return datensatz;
    }

    private <T> T konvertiereObjekt(SystemObject objekt) {
        return (T) objektFactory.gibSystemobjekt(objekt);
    }

    private Regelbasis konvertiereRegelbasis(ResultData result) {
        if (!result.hasData()) {
            String message = MessageFormat.format(
                    "Parameter für Regelbasis {1} am Objekt {0} nicht versorgt, Inferenz findet nicht statt.",
                    result.getObject(),
                    result.getDataDescription().getAttributeGroup()
            );
            log.warning(message);
            messageSender.sendMessage(
                    result.getObject().getPid(),
                    MessageType.SYSTEM_DOMAIN,
                    null,
                    MessageGrade.WARNING,
                    result.getObject(),
                    MessageState.MESSAGE,
                    null,
                    message
            );
            return new Regelbasis();
        }

        Data data = result.getData();
        List<Regel> regeln = new ArrayList<>();
        Data.Array regelnArray = data.getArray("Regeln");
        for (int i = 0; i < regelnArray.getLength(); i++) {
            Data regelItem = regelnArray.getItem(i);
            regeln.add(konvertiereRegel(regelItem));
        }
        return new Regelbasis(regeln);
    }

    private static Regel konvertiereRegel(Data regel) {
        List<Ausdruck<Zugehoerigkeit>> praemisse = bestimmePraemisse(regel);
        TermVariable konklusion = bestimmeKonklusion(regel);
        Sicherheitsgrad sicherheitsgrad = bestimmeSicherheitsgrad(regel);
        return Regel.wenn(new Minimum(praemisse)).dann(konklusion).mitSicherheitsgrad(sicherheitsgrad).erzeuge();
    }

    private static List<Ausdruck<Zugehoerigkeit>> bestimmePraemisse(Data regel) {
        List<Ausdruck<Zugehoerigkeit>> praemisse = new ArrayList<>();
        Data.Array praemisseArray = regel.getArray("Prämisse");
        for (int i = 0; i < praemisseArray.getLength(); i++) {
            Data praemisseItem = praemisseArray.getItem(i);
            praemisse.add(konvertiereFuzzyVariable(praemisseItem));
        }
        return praemisse;
    }

    private static TermVariable bestimmeKonklusion(Data regel) {
        String fuzzyVariable = regel.getItem("Konklusion").getTextValue("Variable").getText();
        String term = regel.getItem("Konklusion").getTextValue("Term").getText();
        return new TermVariable(fuzzyVariable, term);
    }

    private static Sicherheitsgrad bestimmeSicherheitsgrad(Data regel) {
        return Sicherheitsgrad.von(regel.getScaledValue("Sicherheitsgrad").doubleValue());
    }

    private static Ausdruck<Zugehoerigkeit> konvertiereFuzzyVariable(Data variable) {
        String fuzzyVariable;
        int zyklus = variable.getUnscaledValue("Zyklus").intValue();
        if (zyklus > 0) {
            fuzzyVariable = variable.getTextValue("Variable").getText() + zyklus;
        } else {
            fuzzyVariable = variable.getTextValue("Variable").getText();
        }
        String[] termnamen = variable.getTextArray("Term").getTextArray();
        if (termnamen.length == 1) {
            return new TermVariable(fuzzyVariable, termnamen[0]);
        } else {
            List<TermVariable> terme = Stream.of(termnamen)
                    .map(t -> new TermVariable(fuzzyVariable, t))
                    .collect(Collectors.toList());
            return new Maximum(terme);
        }
    }

    Event<Datensatz<?, Regelbasis>> regelbasisEmpfangen() {
        return regelbasisEmpfangen;
    }

    private class Receiver implements ClientReceiverInterface {

        @Override
        public void update(ResultData[] results) {
            for (ResultData r : results) {
                regelbasisEmpfangen.send(konvertiereDatensatz(r));
            }
        }

    }

}
