/*
 * Segment 5 Intelligente Analyseverfahren, SW-Einheit 5.2 Straßensubsegmentanalyse
 * Copyright (C) 2007-2018 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.subsegana.subsegana.portals;

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.iav.fuzzylib.fuzzylib.*;
import de.bsvrz.iav.subsegana.subsegana.domain.shared.*;
import de.bsvrz.iav.subsegana.subsegana.domain.umfelddaten.*;
import de.bsvrz.iav.subsegana.subsegana.domain.verkehrsdaten.*;
import de.bsvrz.sys.funclib.debug.*;
import de.bsvrz.sys.funclib.operatingMessage.*;

import java.text.*;
import java.util.*;

import static de.bsvrz.iav.subsegana.subsegana.domain.umfelddaten.Umfeldgroessen.*;
import static de.bsvrz.iav.subsegana.subsegana.domain.verkehrsdaten.Verkehrsgroessen.*;

/**
 * Sendet Fuzzy-Werte an den Datenverteiler.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
@SuppressWarnings("ClassFanOutComplexity")
public class FuzzyWertSender extends Kommunikation {

    private static final String ATG_ANTEIL_LKW = "atg.anteilLkw";
    private static final String ATG_BEMESSUNGSDICHTE = "atg.bemessungsDichte";
    private static final String ATG_GESCHWINDIGKEIT = "atg.verkehrGeschwindigkeit";
    private static final String ATG_VERKEHRSSTAERKE_STUNDE = "atg.verkehrsStärkeStunde";
    private static final String ATG_VERKEHRSUNRUHE = "atg.verkehrsUnruhe";
    private static final String ATG_VERKEHRSZUSTAND = "atg.verkehrsZustand";
    private static final String ATG_VERKEHRSZUSTAND_UNRUHE = "atg.verkehrsZustandUnruhe";

    private static final String ATG_NIEDERSCHLAGSINTENSITAET = "atg.ufdsNiederschlagsIntensitätFuzzy";
    private static final String ATG_WINDGESCHWINDIGKEIT = "atg.ufdsWindGeschwindigkeitMittelWertFuzzy";
    private static final String ATG_SICHTWEITE = "atg.ufdsSichtWeiteFuzzy";
    private static final String ATG_HELLIGKEIT = "atg.ufdsHelligkeitFuzzy";
    private static final String ATG_WINDRICHTUNG = "atg.ufdsWindRichtungFuzzy";
    private static final String ATG_GLAETTE = "atg.ufdmsGlätteFuzzy";
    private static final String ATG_SICHTVERHAELTNISSE = "atg.ufdmsSichtVerhältnisseFuzzy";
    private static final String ATG_FAHRBAHNZUSTAND = "atg.ufdmsFahrBahnZustandFuzzy";

    private static final String ASP_PID = "asp.fuzzy";

    private final Debug log = Debug.getLogger();
    private final MessageSender messageSender = MessageSender.getInstance();
    private final ClientSenderInterface sender = new SenderOhneSendesteuerung();

    public FuzzyWertSender(ClientDavInterface verbindung) {
        super(verbindung);
    }

    public void anmelden(Collection<Systemobjekt> objekte) {
        meldeVerkehrsdatenAn(objekte);
        meldeUmfelddatenAn(objekte);
    }

    private void meldeVerkehrsdatenAn(Collection<Systemobjekt> objekte) {
        Collection<SystemObject> messquerschnitte = konvertiereObjekte(objekte, obj -> obj instanceof Messquerschnitt);
        anmelden(messquerschnitte, ATG_ANTEIL_LKW);
        anmelden(messquerschnitte, ATG_BEMESSUNGSDICHTE);
        anmelden(messquerschnitte, ATG_GESCHWINDIGKEIT);
        anmelden(messquerschnitte, ATG_VERKEHRSSTAERKE_STUNDE);
        anmelden(messquerschnitte, ATG_VERKEHRSUNRUHE);
        anmelden(messquerschnitte, ATG_VERKEHRSZUSTAND);
        anmelden(messquerschnitte, ATG_VERKEHRSZUSTAND_UNRUHE);
        log.config(MessageFormat.format("Als Sender von Fuzzy-Werten für {0} Messquerschnitte angemeldet.", messquerschnitte.size()));

        Collection<SystemObject> strassenteilsegmente = konvertiereObjekte(objekte, obj -> obj instanceof Strassenteilsegment);
        anmelden(strassenteilsegmente, ATG_ANTEIL_LKW);
        anmelden(strassenteilsegmente, ATG_BEMESSUNGSDICHTE);
        anmelden(strassenteilsegmente, ATG_GESCHWINDIGKEIT);
        anmelden(strassenteilsegmente, ATG_VERKEHRSSTAERKE_STUNDE);
        anmelden(strassenteilsegmente, ATG_VERKEHRSUNRUHE);
        anmelden(strassenteilsegmente, ATG_VERKEHRSZUSTAND);
        anmelden(strassenteilsegmente, ATG_VERKEHRSZUSTAND_UNRUHE);
        log.config(MessageFormat.format("Als Sender von Fuzzy-Werten für {0} Straßenteilsegmenten angemeldet.", strassenteilsegmente.size()));
    }

    private void meldeUmfelddatenAn(Collection<Systemobjekt> objekte) {
        Collection<SystemObject> umfelddatenmessstellen = konvertiereObjekte(objekte, obj -> obj instanceof Umfelddatenmessstelle);
        anmelden(umfelddatenmessstellen, ATG_GLAETTE);
        anmelden(umfelddatenmessstellen, ATG_SICHTVERHAELTNISSE);
        anmelden(umfelddatenmessstellen, ATG_FAHRBAHNZUSTAND);
        log.config(MessageFormat.format("Als Sender von Fuzzy-Werten für {0} Umfelddatenmessstellen angemeldet.", umfelddatenmessstellen.size()));

        Collection<SystemObject> helligkeitssensoren = konvertiereObjekte(objekte, obj -> obj instanceof Helligkeitssensor);
        anmelden(helligkeitssensoren, ATG_HELLIGKEIT);
        log.config(MessageFormat.format("Als Sender von Fuzzy-Werten für {0} Helligkeitssensoren angemeldet.", helligkeitssensoren.size()));

        Collection<SystemObject> niederschlagsintensitaetssensoren = konvertiereObjekte(objekte, obj -> obj instanceof Niederschlagsintensitaetssensor);
        anmelden(niederschlagsintensitaetssensoren, ATG_NIEDERSCHLAGSINTENSITAET);
        log.config(MessageFormat.format("Als Sender von Fuzzy-Werten für {0} Niederschlagsintentensitätssensoren angemeldet.", niederschlagsintensitaetssensoren.size()));

        Collection<SystemObject> sichtweitesensoren = konvertiereObjekte(objekte, obj -> obj instanceof Sichtweitesensor);
        anmelden(sichtweitesensoren, ATG_SICHTWEITE);
        log.config(MessageFormat.format("Als Sender von Fuzzy-Werten für {0} Sichtweitesensoren angemeldet.", sichtweitesensoren.size()));

        Collection<SystemObject> windgeschwindigkeitssensoren = konvertiereObjekte(objekte, obj -> obj instanceof Windgeschwindigkeitssensor);
        anmelden(windgeschwindigkeitssensoren, ATG_WINDGESCHWINDIGKEIT);
        log.config(MessageFormat.format("Als Sender von Fuzzy-Werten für {0} Windgeschwindigkeitssensoren angemeldet.", windgeschwindigkeitssensoren.size()));

        Collection<SystemObject> windrichtungssensoren = konvertiereObjekte(objekte, obj -> obj instanceof Windrichtungssensor);
        anmelden(windrichtungssensoren, ATG_WINDRICHTUNG);
        log.config(MessageFormat.format("Als Sender von Fuzzy-Werten für {0} Windrichtungssensoren angemeldet.", windrichtungssensoren.size()));
    }

    private void anmelden(Collection<SystemObject> objects, String attributgruppePid) {
        DataDescription dd = datenbeschreibung(attributgruppePid, ASP_PID);
        try {
            getVerbindung().subscribeSender(sender, objects, dd, SenderRole.source());
        } catch (OneSubscriptionPerSendData e) {
            log.error("Sendeanmeldung fehlgeschlagen: " + e, Arrays.asList(objects, dd));
            throw new IllegalStateException("Sendeanmeldung fehlgeschlagen", e);
        }
    }

    public void sende(Collection<Datensatz<?, FuzzyVariable>> datensaetze) {
        ResultData[] results = datensaetze.stream()
                .map(this::konvertiereDatensatz)
                .toArray(ResultData[]::new);
        try {
            getVerbindung().sendData(results);
            log.fine(MessageFormat.format("{0} Fuzzy-Werte gesendet", datensaetze.size()));
        } catch (SendSubscriptionNotConfirmed e) {
            String message = "Fuzzy-Werte konnten nicht gesendet werden: " + e;
            log.warning(message, e);
            messageSender.sendMessage(MessageType.SYSTEM_DOMAIN, MessageGrade.WARNING, message);
        }
    }

    private ResultData konvertiereDatensatz(Datensatz<?, FuzzyVariable> datensatz) {
        SystemObject obj = getModell().getObject(datensatz.getSystemobjekt().getPid());
        AttributeGroup atg = bestimmeAttributgruppe(datensatz.getDatum());
        long time = konvertiereZeitstempel(datensatz.getZeitstempel());
        Data data = konvertiereDatensatz(datensatz, getVerbindung().createData(atg));
        return new ResultData(obj, datenbeschreibung(atg.getPid(), ASP_PID), time, data);
    }

    @SuppressWarnings({"CyclomaticComplexity", "JavaNCSS"})
    private AttributeGroup bestimmeAttributgruppe(FuzzyVariable fuzzyWert) {
        String pid;
        switch (fuzzyWert.getName()) {
            case QKFZ:
                pid = ATG_VERKEHRSSTAERKE_STUNDE;
                break;
            case VKFZ:
                pid = ATG_GESCHWINDIGKEIT;
                break;
            case SKFZ:
                pid = ATG_VERKEHRSUNRUHE;
                break;
            case ALKW:
                pid = ATG_ANTEIL_LKW;
                break;
            case KB:
                pid = ATG_BEMESSUNGSDICHTE;
                break;
            case VERKEHRSZUSTAND:
                pid = ATG_VERKEHRSZUSTAND;
                break;
            case VERKEHRSZUSTAND_UNRUHE:
                pid = ATG_VERKEHRSZUSTAND_UNRUHE;
                break;
            case GLAETTE:
                pid = ATG_GLAETTE;
                break;
            case HELLIGKEIT:
                pid = ATG_HELLIGKEIT;
                break;
            case NIEDERSCHLAGSINTENSITAET:
                pid = ATG_NIEDERSCHLAGSINTENSITAET;
                break;
            case SICHTWEITE:
                pid = ATG_SICHTWEITE;
                break;
            case WINDGESCHWINDIGKEIT:
                pid = ATG_WINDGESCHWINDIGKEIT;
                break;
            case WINDRICHTUNG:
                pid = ATG_WINDRICHTUNG;
                break;
            case SICHTVERHAELTNISSE:
                pid = ATG_SICHTVERHAELTNISSE;
                break;
            case FAHRBAHNZUSTAND:
                pid = ATG_FAHRBAHNZUSTAND;
                break;
            default:
                throw new IllegalStateException("[unreachable code] Nicht behandelter Fuzzy-Variable: " + fuzzyWert.getName());
        }
        return getModell().getAttributeGroup(pid);
    }

    private Data konvertiereDatensatz(Datensatz<?, FuzzyVariable> datensatz, Data data) {
        FuzzyVariable fuzzyVariable = datensatz.getDatum();
        Data.Array array = data.getArray(fuzzyVariable.getName());
        array.setLength(fuzzyVariable.getTerme().size());
        for (int i = 0; i < array.getLength(); i++) {
            Term term = fuzzyVariable.getTerme().get(i);
            array.getItem(i).getTextValue("Name").setText(term.getName());
            if (term.getZugehoerigkeit().nonNull()) {
                array.getItem(i).getScaledValue("Zugehörigkeit").set(term.getZugehoerigkeit().getWert());
            } else {
                array.getItem(i).getUnscaledValue("Zugehörigkeit").setText("nicht ermittelbar");
            }
        }
        return data;
    }

}
